/*
chil.c

A set of C functions to read all Chilbolton-format radar data since 1993

The reason I wrote this is that there have been many slight changes
to the format in recent years, particularly with the cloud radars. It
provides a programming interface such that the programs themselves do
not need to know the details of how to decode the header, read the
ray data etc. The length of the CHILReadInformationRecord() function
illustrates how overburdened this format has become.

These functions are not guaranteed to work with all Chilbolton-format
data, or any at all for that matter. In particular I've never tried
it on Singapore data, and it may have problems with older CAMRa data
and some of the earlier cloud radar data. 

R J Hogan 

Project started 1/6/99
Last update 20/10/99
*/


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glob.h>
#include "chil.h"

#ifdef MATLAB
#include <mex.h>
#endif

#if __i386__ || __386__
#define __IS_I386_ARCH__
#endif


/* Define some arrays of strings  */
/*
char *field_names[] = { UNKNOWN_NAME , ZH_NAME , ZDR_NAME , 
LDR_NAME , PHIDP_NAME , VEL1_NAME , VEL2_NAME , WIDTH_NAME };
char *Zh2_name=ZH2_NAME;
char *system_names[] = { "unknown","camra","rabelais","galileo","singapore"};
*/
/* Data buffers to hold data */

#define RAY_BUFFER_SIZE 4096
short int ray_buffer[RAY_BUFFER_SIZE],tmp_buffer[RAY_BUFFER_SIZE];
unsigned char pad_block[BLKSIZE];
char secondary_sum_dir[MAX_FILENAME_LENGTH]="\0";

int CHILGetFileList(uword day,uword month,uword year,int Argc,char **Argv,FileList * filelist) {
int i,status;
glob_t globbuf;
FILE *sumfile;
uword raster,yraster=0,pos=0;
uword tday,tmonth,tyear;
uword yday,ymonth,yyear;
unsigned int GLOB_FLAG=0;
char *lastfilename=NULL;
char tmp[255];

filelist->n=0;

#ifdef DEBUG
fprintf(stderr,"Globbing the file list\n");
#endif
/* First we want to glob the argument list (stopping if we reach one that begins
with "-") to obtain a list of filenames */
for (i=0;i<Argc;i++) {
   if (*Argv[i] == '-')
      break;
   if ((status=glob(Argv[i],GLOB_FLAG,NULL,&globbuf)) == GLOB_NOSPACE) {
      globfree(&globbuf);
      return -1;
   }
   GLOB_FLAG=GLOB_APPEND;
}

#ifdef DEBUG
fprintf(stderr,"Getting yesterdays date\n");
#endif
/* get date for yesterday */
yday=day;ymonth=month;yyear=year;
CHILDecrDate(&yday,&ymonth,&yyear);
#ifdef DEBUG
fprintf(stderr,"Searching summary files\n");
#endif

status=0;
/* Loop through all matched filenames, search corresponding summary files for the date required */
for (i=0;i<globbuf.gl_pathc;i++) {
   if ((sumfile=CHILOpenSummary(globbuf.gl_pathv[i])) != NULL) {
#ifdef DEBUG
fprintf(stderr,"Openned %s\n",globbuf.gl_pathv[i]);
#endif
      fgets(tmp,255,sumfile); /* skip first line */
      while(fgets(tmp, 255, sumfile) != '\0') {
         sscanf(tmp, "%*5d %3hu %*8c %2hu/%2hu/%2hu",&raster,&tday,&tmonth,&tyear);
#ifdef DEBUG
fprintf(stderr,"Read a line\n");
#endif
/* Decide if raster is for yesterday, today or neither */
         if (status==0 && tday==yday && tmonth==ymonth && tyear==yyear) {
            yraster=raster;
         }
         else if (tday==day && tmonth==month && tyear==year) {
            if (status==0 && yraster > 0) {
               strncpy(filelist->name[0]=filelist->buffer,lastfilename,MAX_FILENAME_LENGTH);
               pos=strlen(lastfilename)+1;
               filelist->first[0]=filelist->last[0]=yraster;
               (filelist->n)=1;
               status=1;
               yraster=0;
            }
            if (lastfilename != globbuf.gl_pathv[i]) {
               if (filelist->n >= MAX_FILES) {
                  status=2;
                  break;
               }
               strncpy(filelist->name[filelist->n]=filelist->buffer+pos,globbuf.gl_pathv[i],MAX_FILENAME_LENGTH);
               filelist->first[filelist->n]=filelist->last[filelist->n]=raster;
               pos+=strlen(globbuf.gl_pathv[i])+1;
               ++filelist->n;
            }
            else {
               filelist->last[filelist->n-1]=raster;
            }
         }
         else if (status==1) {
            if (yraster > 0) {
               strncpy(filelist->name[0]=filelist->buffer,lastfilename,MAX_FILENAME_LENGTH);
               filelist->first[0]=filelist->last[0]=yraster;
               (filelist->n)=1;
            }
            status=2;
            break;
         }
         lastfilename=globbuf.gl_pathv[i];
      }
      fclose(sumfile);
      if (status==2)
         break;
   }
}
#ifdef DEBUG
fprintf(stderr,"Freeing list\n");
#endif
globfree(&globbuf);
return filelist->n;
}
FILE *CHILOpenSummary(char * datfilename) {
   char sumfilename[MAX_FILENAME_LENGTH];
   char *dot,*slash;
   FILE *sumfile=NULL;

   strcpy(sumfilename,datfilename);
   if ((dot=strrchr(sumfilename,'.')) != '\0') {
      if (strncmp(dot+1,"dat",3) == 0) {
         strncpy(dot+1,"sum",3);
         if ((sumfile=fopen(sumfilename,"r")) == NULL) {
/* Summary file not found in the same directory as the data... 
try contents of secondary_sum_dir */
            if (secondary_sum_dir[0]=='\0')
               return NULL;
            strncpy(sumfilename,secondary_sum_dir,MAX_FILENAME_LENGTH);
            if ((slash=strrchr(datfilename,'/')) !='\0') {
               strcpy(sumfilename+strlen(sumfilename),slash);
            }
            else {
               strcpy(sumfilename+strlen(sumfilename),"/");
               strcpy(sumfilename+strlen(sumfilename),datfilename);
            }
            if ((dot=strrchr(sumfilename,'.')) != '\0') {
               if (strncmp(dot+1,"dat",3) == 0) {
                  strncpy(dot+1,"sum",3);
                  return (fopen(sumfilename,"r"));
               }
            }
         }
      }
   }
   return sumfile;
}
int CHILOpen(char * filename, ChilboltonFile * chilfile) {
   char *dot,*slash;
   char sumfilename[MAX_FILENAME_LENGTH];
   char tmp[255],leaveloop=0;
   FILE *sumfile;

   if ((chilfile->file=fopen(filename,"r"))==NULL) {
      chilfile->position = FILE_CLOSED;
      return FILE_CLOSED;
   }
   chilfile->nrasters=0;
   chilfile->summary_exists=0;
   chilfile->bytes_read=0;
   chilfile->is_raw=0;
   strncpy(chilfile->name,filename,MAX_FILENAME_LENGTH);
   chilfile->name[MAX_FILENAME_LENGTH-1]='\0';
   if ((dot=strrchr(filename,'.')) != '\0') {
      if (strncmp(dot+1,"raw",3) == 0)
         chilfile->is_raw = 1;
      else {
         if ((sumfile=CHILOpenSummary(filename)) != NULL) {
            chilfile->summary_exists=1;
            fgets(tmp,255,sumfile); /* skip first line */
            while(fgets(tmp, 255, sumfile) != '\0') {
//               sscanf(tmp, "%hu %hu", (chilfile->block)+(chilfile->nrasters),(chilfile->raster)+(chilfile->nrasters));
               sscanf(tmp, "%5hu %3hu %*8c %2hu/%2hu/%2hu",(chilfile->block)+(chilfile->nrasters),(chilfile->raster)+(chilfile->nrasters),(chilfile->day)+(chilfile->nrasters),(chilfile->month)+(chilfile->nrasters),(chilfile->year)+(chilfile->nrasters));
               (chilfile->nrasters)++;
            }
            fclose(sumfile);
         }
      }
   }
   chilfile->position = START_OF_FILE;
   return START_OF_FILE;
}

int CHILClose(ChilboltonFile * chilfile) {
   chilfile->position=FILE_CLOSED;
   return (fclose(chilfile->file));
}

char *CHILSecondarySumDir(char * sec_sum_dir) {
   return (strncpy(secondary_sum_dir,sec_sum_dir,MAX_FILENAME_LENGTH));
}
int CHILSeekRaster(ChilboltonFile * chilfile,uword required_raster) {
   unsigned short int i=0;
   if (! chilfile->summary_exists) {
      fprintf(stderr,"Summary file does not exist so cannot seek rasters\n");
      return 1;
   }
   else {
      while (chilfile->raster[i] != required_raster && chilfile->nrasters > i++);
      if (chilfile->raster[i]==required_raster) {
         fseek(chilfile->file,BLKSIZE*(chilfile->block[i]-1),SEEK_SET);
         chilfile->position=START_OF_IR;
         chilfile->iraster=i;
         chilfile->iscan=0;
         chilfile->bytes_read=0;
         return 0;
      }
      else {
//         fprintf(stderr,"Cannot find raster number %hu in summary file\n",required_raster);
         return 2;
      }   
   }
};

int CHILSkipScans(ChilboltonFile * chilfile,RasterInfo * info,int nscans) {
int iscan=0,ir=0;
if (nscans>0) {
   ir = (chilfile->position == START_OF_IR || chilfile->position == START_OF_FILE);
//   fprintf(stderr,"**%d**\n",ir);
   do {
      if (ir) {
//         fprintf(stderr,"Skipping header...\n");
         if (fread(tmp_buffer,1,64,chilfile->file) < 64) {
            fprintf(stderr,"End of file reached\n");
            return FILE_CLOSED;
         }
         (chilfile->bytes_read) +=64;
         iscan++;
         (chilfile->iscan)++;
         chilfile->position = START_OF_DATA;
      }
      else
         ir=1;
      while (CHILSkipRay(chilfile,info)==START_OF_RAY);
   } while (iscan < nscans && chilfile->iscan > 0);
}
if (chilfile->iscan == 0) {
   return 1;
}
return 0;
}
int CHILSkipRay(ChilboltonFile * chilfile,RasterInfo * info) {
   unsigned char signal[2];
   unsigned short nwords;

if (chilfile->position !=START_OF_RAY && chilfile->position !=START_OF_DATA) {
   fprintf(stderr,"Attempt to skip over a ray when at an inappropriate position in file\n");
   return UNKNOWN_POSITION;
}
else {
   fread(signal,sizeof(char),2,chilfile->file);
   (chilfile->bytes_read) += 2;
   nwords=signal[1] | signal[0] << 8;
   switch (nwords) {
      case END_OF_SCAN:
         return (chilfile->position=START_OF_IR);
         break;
      case END_OF_RASTER:case RASTER_ABORTED:
         return (CHILNextRaster(chilfile,info));
         break;
      default:
//        fprintf(stderr,"%hu ",nwords);
        fseek(chilfile->file,(nwords-1)*2,SEEK_CUR);
//        fread(&tmp_buffer,sizeof(short),nwords,chilfile->file);
         (chilfile->bytes_read) += (nwords-1)*2;
         return (chilfile->position=START_OF_RAY);
   }
}
}



int CHILReadInformationRecord(ChilboltonFile *chilfile,RasterInfo *info) {
#define mw( a, b ) ( a<<8 | b )

   unsigned char *ir, new_ir_format;
   short int i;

   if (chilfile->position != START_OF_IR && chilfile->position != START_OF_FILE ) {
      fprintf(stderr,"Warning: attempt to read Information Record when not at appropriate position\n");
//      CHILPrintFileInformation(chilfile);
      return FILE_CLOSED;
   }

//   if (chilfile->position == START_OF_FILE) 
//      chilfile->bytes_read = 0;
   if (i=fread(info->raw,sizeof(info->raw),1,chilfile->file) < 1) {
      return FILE_CLOSED;
   }
   (chilfile->bytes_read) += sizeof(info->raw);
   chilfile->position = START_OF_DATA;
   ir=info->raw;

/* First fill in the obvious fields */
   info->recording_flag = ir[1];
   info->min_gate = mw(ir[9], ir[10]);
   info->max_gate = mw(ir[11], ir[12]);
//   info->RHI_display_max = ir[15];
   info->threshold = mw(ir[16], ir[17]);
   info->min_angle = mw(ir[20], ir[21]);
   info->max_angle = mw(ir[22], ir[23]);
   info->central = mw(ir[24], ir[25]);
   info->system_id = ir[29]; /* This is 0 in the old format.. set correctly below */
   info->day = ir[31];
   info->month = ir[32];
   info->year = ir[33];
   info->nscans = ir[34];
   info->expt_id = ir[35];
   info->scan_velocity = ir[36];
   info->scan_type = ir[37];
   info->raster = mw(ir[39], ir[40]);
   info->sens_flag = ir[41];
   info->hour = ir[42];
   info->operator[0] = ir[43];
   info->operator[1] = ir[44];
   info->operator[2] = ir[45];
   info->minute = ir[46];
   info->second = ir[47];
   info->init_angle = mw(ir[48], ir[49]);
   info->angle_inc = mw(ir[50], ir[51]);
   info->file = mw(ir[52], ir[53]);
   info->time_stamping = ir[58];
   info->cmpl_scans = ir[59];

   if (ir[57] == 1) 
      info->range_normalised = 0;
   else
      info->range_normalised = 1;

//fprintf(stderr,"ID byte: %x\n",ir[0]);
/* Now the tricky bit: find out what the system is and what parameters are stored */
    if (info->scan_type == FIXED_93 && ir[4] == 0)
      info->cloud_data=1;
   else
      info->cloud_data=0;
   info->nfields=0;
   new_ir_format=info->system_id == 0 ? 0 : 1;

//   fprintf(stderr,"New IR format? %hu\n",new_ir_format);

   if (new_ir_format) {
      /* NEW IR FORMAT (as of spring 1999) */
      for (i=0;i<7;i++) {
         if (ir[27] & (1L<<i)) {
            info->field_type[info->nfields++] = i+1;
         }
         if (ir[27] & (1L<<7)) {
            for (i=0;i<8;i++) {
               if (ir[28] & (1L<<i)) {
                  info->field_type[info->nfields++] = i+8;
               }
            }
         }
      }         
      /* Now get field_ID from field_type */
//      for (i=0;i<info->nfields;i++) {
//         info->field_id[i]=info->system_id
//      }

   } else {
      /* OLD IR FORMAT */
      for (i=26;i<29;i++) {
         if (ir[i])
            info->field_id[info->nfields++] = ir[i];
      }
      /* Non-foolproof way to guess the system: */
      if (info->field_id[0] < 30) {
         info->system_id = CAMRA;
      } else if (info->field_id[0] < 40 ) {
         info->system_id = RABELAIS;
      } else if (info->field_id[0] < 50 ) {
         info->system_id = SINGAPORE;
      } else if (info->field_id[0] < 60 ) {
         info->system_id = GALILEO;
      } else {
         info->system_id = UNKNOWN_SYSTEM;
      }
      /* Get field type from field ID */
      for (i=0;i<info->nfields;i++) {
//fprintf(stderr,"%hd\n",i);
         switch (info->field_id[i]) {
            case CAMRA_ZH:case RABELAIS_ZH:case SINGAPORE_ZH:case GALILEO_ZH:
               info->field_type[i]=ZH;
               break;
            case CAMRA_ZDR:case RABELAIS_ZDR:
               info->field_type[i]=ZDR;
               break;
            case CAMRA_LDR:case SINGAPORE_LDR:
               info->field_type[i]=LDR;
               break;
            case CAMRA_PHIDP:case RABELAIS_PHIDP:
               info->field_type[i]=PHIDP;
               break;
            case CAMRA_V:case CAMRA_VH:case RABELAIS_V:case SINGAPORE_V:case GALILEO_V:
               info->field_type[i]=V1;
               break;
            case CAMRA_VV:case CAMRA_V2:case RABELAIS_VV:
               info->field_type[i]=V2;
               break;
            case CAMRA_WIDTH:case SINGAPORE_WIDTH:
               info->field_type[i]=WIDTH_H;
               info->field_type[++i]=WIDTH_V;
               break;
            default:
               info->field_type[i]=UNKNOWN;
         }
      }
   }

/* Get frequency from system id */
   switch (info->system_id) {
      case CAMRA:
         info->frequency=CAMRA_FREQ;
         break;
      case RABELAIS:
         info->frequency=RABELAIS_FREQ;
         break;
      case GALILEO:
         info->frequency=GALILEO_FREQ;
         break;
      case SINGAPORE:
         info->frequency=SINGAPORE_FREQ;
         break;
      default:
         info->frequency=0.0;
   } 

/* Get the gate spacing and range offset*/
   info->range_offset=0;
   if (info->cloud_data) {
      if (new_ir_format) {
         info->gate_spacing = mw(ir[60], ir[61]);
         info->range_offset = -(255 - ir[62]) * 256 - (256 - ir[63]);
//fprintf(stderr,"gate spacing: %hd, range offset: %hd\n",info->gate_spacing,info->range_offset);
      }
      else {
         switch (info->system_id) {
            case RABELAIS:case SINGAPORE:
               info->gate_spacing=ir[14]*MIN_SPACING;
               break;
            case GALILEO:
               if (info->file < 8282)
                  info->gate_spacing=ir[14]*MIN_SPACING;
               else
                  info->gate_spacing=ir[14]*MIN_SPACING_GALILEO_NEW;
               break;
         }
      }
   }
   else {
//      if (new_ir_format) {
         switch (info->system_id) {
            case CAMRA:
               switch (info->field_type[0]) {
                  case ZH:case ZDR:case LDR:
                     info->gate_spacing=ir[4]*ir[13]*MIN_SPACING;
                     break;
                  case V1:case V2:case PHIDP:
                     info->gate_spacing=ir[3]*MIN_SPACING;
                     break;
               }
            case RABELAIS:
               info->gate_spacing=ir[7]*ir[8]*MIN_SPACING;
               break;
            case GALILEO:
               if (info->file < 6000)
//                  info->gate_spacing=ir[38]*ir[14]*MIN_SPACING;
                  info->gate_spacing=ir[7]*ir[8]*MIN_SPACING;
               else
                  info->gate_spacing=ir[7]*ir[8]*MIN_SPACING_GALILEO_NEW;
               break;
         }
//      }
   }
   if (info->gate_spacing == 0) {
      fprintf(stderr,"Warning: could not identify range spacing\n");
//      CHILPrintInformationRecord(info);
   }


//   info->spacing_raw=
//   info->spacing_ave=
//   info->time_stamping

   for (i=0;i<info->nfields;i++) {
      switch (info->field_type[i]) {
         case ZH:
            if (info->system_id == CAMRA) {
               info->field_function[i]=CHILzh;
            }
            else {
               info->field_function[i]=CHILzh_highprecision;
            }
            break;
         case ZDR:
            if (info->system_id == CAMRA) {
               info->field_function[i]=CHILzdr;
            }
            else {
               info->field_function[i]=CHILzdr_highprecision;
            }
            break;
          case LDR:
            info->field_function[i]=CHILldr;
            break;
         case PHIDP:
            info->field_function[i]=CHILphidp;
            break;
         case V1:case V2:case WIDTH:case WIDTH_H:case WIDTH_V
            info->field_function[i]=CHILvel;
            break;
         default:
            info->field_function[i]=CHILraw;
      }
   }

/* Universal format sweep modes */
    switch (info->scan_type) {
    case RHI_93:
	info->sweep_mode = UF_RHI;
	break;
    case PPI_93:
	info->sweep_mode = UF_PPI;
	break;
    case SP_93:
	info->sweep_mode = UF_COPLANE;
	break;
    case FIXED_93:
	info->sweep_mode = UF_DWELL;
	break;
    case CAL_93:
	info->sweep_mode = UF_CAL;
	break;
    case TS_93:
	info->sweep_mode = UF_MANUAL;
	break;
    default:
	fprintf(stderr, "Impossible sweep mode: aborting\n");
	exit(-3);
    }

/* often in large multi-day cloud radar files the date in the rasters is wrong - use summary files */
   if (info->cloud_data == 1 && chilfile->summary_exists) {
      for (i=0;i<chilfile->nrasters;i++) {
         if (chilfile->raster[i]==info->raster) {
#ifdef DEBUG
fprintf(stderr,"Fixing date\n");
#endif
            info->day=chilfile->day[i];
            info->month=chilfile->month[i];
            info->year=chilfile->year[i];
            break;
         }
      }
   }
   return START_OF_DATA;
}
char *CHILFieldName(const Ray *ray,int field,char * name) {
char *field_names[] = { UNKNOWN_NAME , ZH_NAME , ZDR_NAME , 
LDR_NAME , PHIDP_NAME , VEL1_NAME , VEL2_NAME , WIDTH_H_NAME, WIDTH_V_NAME};
/* Find if v and Zh are unque */
unsigned char i,sum=0,n=0;
switch (ray->field_type[field]) {
   case ZH:
      for (i=0;i < ray->nfields;i++) {
         if (ray->field_type[i] == ZH) {
            sum++;
            if (i==field) n=sum;
         }
      }
      if (sum < 2 || sum==n)
         return (strcpy(name,ZH_NAME));
      else
         return (strcpy(name,ZH2_NAME));
      break;
   case V1:case V2:
       for (i=0;i < ray->nfields;i++) {
         if (ray->field_type[i] == V1 || ray->field_type[i] == V2 ) {
            sum++;
            if (i==field) n=sum;
         }
      }
      if (sum < 2)
         return (strcpy(name,VEL_NAME));
      else if (sum==n)
         return (strcpy(name,VEL2_NAME));
      else
         return (strcpy(name,VEL1_NAME));
      break;
   default:
      return (strcpy(name,field_names[ray->field_type[field]]));
}
}
char *CHILFieldUnits(const Ray *ray,int field,char * units) {
char *field_units[] = { UNKNOWN_UNITS , ZH_UNITS , ZDR_UNITS , 
LDR_UNITS , PHIDP_UNITS , VEL1_UNITS , VEL2_UNITS , WIDTH_UNITS ,
WIDTH_UNITS, UNKNOWN_UNITS , UNKNOWN_UNITS , 
UNKNOWN_UNITS , UNKNOWN_UNITS , UNKNOWN_UNITS , 
UNKNOWN_UNITS , VELSTD_UNITS};
return (strcpy(units,field_units[ray->field_type[field]]));
}
char *CHILFieldDescription(const Ray *ray,int field,char * desc) {
char *field_desc[] = { "Unknown variable" , "Radar reflectivity factor" , 
"Differential reflectivity" , "Linear depolarisation ratio", "Differential phase shift",
"Doppler velocity (H&V)","Doppler velocity (H)","Doppler spectral width (H)",
"Doppler spectral width (V)", "", "", "", "", "", "", "Doppler velocity standard deviation" };
return (strcpy(desc,field_desc[ray->field_type[field]]));
}
char *CHILSystemName(const RasterInfo *info,char *name) {
char *system_names[] = { "Unknown","CAMRa","Rabelais","Galileo","Singapore"};
return (strcpy(name,system_names[info->system_id]));
}

/* increase a date and account for leap years etc...*/
/* currently Chilbolton format stores the year as a single byte, and one should add
1900 to it to get the correct value. I am assuming it will switch to 100 not 0 in 2000... */
void CHILIncrDate(uword *day,uword *month,uword *year) {
uchar daysinmonth[]={31,28,31,30,31,30,31,31,30,31,30,31};
uword fullyear;
if (*year > 500)
   fullyear=*year;
else
   fullyear=1900+*year;

(*day)++;
if (*day > daysinmonth[*month-1]) {
   /* check for leap year */
   if (!(*day==29 && (fullyear % 4 && !(fullyear % 400)))) {
      *day=1;
      (*month)++;
      if (*month>12) {
         *month=1;
         (*year)++;
      }
   }
}

}
void CHILDecrDate(uword *day,uword *month,uword *year) {
uchar daysinmonth[]={31,28,31,30,31,30,31,31,30,31,30,31};
uword fullyear;

if (*year<500)
   fullyear=1900+*year;
else
   fullyear=*year;

if (--(*day) < 1) {
   if (--(*month) == 2 && (fullyear % 4) && !(fullyear % 400)) {
      /* leap year! */
      *day=29;
   }
   else {
      if (*month<1) {
         (*year)--;
         *month=12;
      }
      *day=daysinmonth[*month-1];
   }
}

}

/* The following is the `internal' function for initialising rays */
/* if `new' is 1 then memory is allocated for the first time, if it is 0 then no allocation takes place */
int CHILSetRay(const RasterInfo *info,int avegates,int averays,Ray *ray, uchar new) {
   long int newsize;
   uchar floatsize,doublesize;
   float *buffer;
   uword i;
   float start_range;

   ray->range_normalised=info->range_normalised;
   ray->correct_range=0;

   ray->utc=0.0;
   ray->last_utc=0.0;
   ray->hour_mod=0;

   ray->sigma_v_floor=0;
   ray->v_floor=0;

   ray->day=info->day;
   ray->month=info->month;
   ray->year=info->year;

   ray->raw_utc=ray->utc;
   ray->raw_day=ray->day;
   ray->raw_month=ray->month;
   ray->raw_year=ray->year;

   ray->averays=averays;
   ray->avegates=avegates;
/*   ray->ngates=(info->max_gate - info->min_gate + 1) / avegates; */
   ray->ngates=(info->max_gate - info->min_gate) / avegates;
   ray->nfields=info->nfields;

/* Work out the number of fields we intend to derive from the raw fields */
/* 
   At the moment the only one we consider is the standard deviation of the 
   raw-ray velocity over the (avegates x averays) pulses we are averaging;
   only implemented if (avegates>1 || averays>1) && raster_info->cloud_data==1
*/
   ray->ntotal=ray->nfields;
   for (i=0;i<ray->nfields && i<MAX_FIELDS;i++) {
      ray->field_type[i]=info->field_type[i];
      if (info->cloud_data==1 && (averays>1 || avegates>1) &&
            (ray->field_type[i]==V1 || ray->field_type[i]==V2)) {
#ifdef DEBUG
   fprintf(stderr,"Setting up sigma-v calculation\n");
#endif
         ray->field_type[ray->ntotal]=V_STD;
         ray->field_info[ray->ntotal]=i;
         ray->ntotal++;
      }
   }

   newsize = ray->ngates * (sizeof(float) * (ray->ntotal+1) + sizeof(double) * ray->ntotal);

   if (new) {
#ifdef DEBUG
      fprintf(stderr,"Allocating space for %ld floats in memory\n",newsize/sizeof(float));
#endif

#ifdef MATLAB
/*      if ((buffer=(float *) mxRealloc(ray->range,newsize)) == NULL) {} */
      if ((buffer=(float *) mxCalloc(newsize/sizeof(float),sizeof(float))) == NULL) {
         mexErrMsgTxt("Cannot allocate memory for ray");
      }
#else
/*      if ((buffer=(float *) realloc(ray->range,newsize)) == NULL) {} */
      if ((buffer=(float *) calloc(newsize/sizeof(float),sizeof(float))) == NULL) {
         fprintf(stderr,"Fatal error: can't allocate memory for rays\n");
         exit(2);
      }
#endif

   ray->maxsize=newsize;
   
   ray->range=(float *) buffer;
   for (i=0;i<ray->ntotal;i++) {
      ray->data[i]=ray->range+ ray->ngates * (i+1);
      ray->sum[i]=((double *) (ray->range + (ray->ntotal+1) * ray->ngates))
            + (ray->ngates * i);
#ifdef DEBUG
fprintf(stderr,"value at 0+%d=%f\n",ray->data[i]-ray->range,*(ray->data[i]));
fprintf(stderr,"value at 0+%d",((float *) ray->sum[i]) - ray->range);
//fprintf(stderr,"=%f\n",(float) *(ray->sum[i]));
#endif
   }
/*   start_range=info->range_offset+(info->min_gate+1)*info->gate_spacing;  */
   start_range=((((float) info->min_gate-1)/((float) avegates))+0.5)*(info->gate_spacing * avegates);
   for (i=0;i<ray->ngates;i++)
      ray->range[i]= (start_range + ((float) i) * avegates * info->gate_spacing) *0.001;
/* Don't need to put the averaging effect into range_offset; 
it's already in start_range*/
   ray->range_offset=info->range_offset /*+0.5*((avegates-1)*info->gate_spacing)*/ ;
   return 0;
   }
   else {
/* Check that new raster has same ray size */
      if (newsize != ray->maxsize) {
         fprintf(stderr,"Warning: re-initialised ray has different size!\n");
         return 1;
      }
   }
   return 0;
}
int CHILInitRay(const RasterInfo *info,int avegates,int averays,Ray *ray) {
return (CHILSetRay(info,avegates,averays,ray,1));
}
int CHILReInitRay(const RasterInfo *info,int avegates,int averays,Ray *ray) {
return (CHILSetRay(info,avegates,averays,ray,0));
}
int CHILFreeRay(Ray * ray) {
#ifdef MATLAB
   mxFree((void *) ray->range);
#else
   free((void *) ray->range);
#endif
}

int CHILReadRawRay(ChilboltonFile * chilfile,RasterInfo * info,Ray * ray) {
   RayHeader header,tmp_header;
   TimeStamp timestamp,tmp_timestamp;
   unsigned char signal[2];
   unsigned short nwords;

/* Read and interpret header */
   fread(signal,sizeof(char),2,chilfile->file);
   (chilfile->bytes_read) += 2;
   nwords = signal[1] | signal[0] << 8;

   if (!(nwords == END_OF_SCAN || nwords == END_OF_RASTER || nwords == RASTER_ABORTED)) {

/* NOTE: this code can read raw data (little endian) as well as processed
(big endian), and also be compiled for either architecture, so we must be careful...*/
#ifdef __IS_I386_ARCH__
#define swap_bytes (! chilfile->is_raw)
#else
#define swap_bytes (chilfile->is_raw)
#endif
      if (swap_bytes) {
         fread(&tmp_header,sizeof(header),1,chilfile->file);
         swab((void *) &tmp_header,(void *) &header,sizeof(header));
      }
      else {
         fread(&header,sizeof(header),1,chilfile->file);
      };
      (chilfile->bytes_read) += sizeof(header);
      nwords -= 3;

      ray->elevation=header.elevation / 60.0;
      ray->azimuth=header.azimuth / 60.0;

      if (info->time_stamping) {
         if (swap_bytes) {
            fread(&tmp_timestamp,sizeof(timestamp),1,chilfile->file);
            swab((void *) &tmp_timestamp,(void *) &timestamp,sizeof(timestamp));
         }
         else {
            fread(&timestamp,sizeof(timestamp),1,chilfile->file);
         }
         (chilfile->bytes_read) += sizeof(timestamp);
         nwords -= 2;
         ray->last_utc=ray->utc;
         ray->utc=((double) timestamp.mins_since_midnight)/60.0 
               + ((double) timestamp.centisecs)/360000.0;
//         fprintf(stderr,"(%d %f %f %hu) ",(int) swap_bytes, ray->utc,ray->last_utc,timestamp.centisecs);
/* The small value is because I've had problems comparing floats that are equal */
         if (ray->utc < ray->last_utc - 0.0003) {
            CHILIncrDate(&(ray->day),&(ray->month),&(ray->year));
         }
      }
      else if (info->cloud_data) {
         /* in the older cloud data the time is encoded into the redundant az and el */
         ray->last_utc=ray->utc;
         ray->utc=info->hour+ray->azimuth + ray->elevation/60.0 + ray->hour_mod;
         if (ray->utc < ray->last_utc - 0.0003) {
            if (ray->utc >=23) {
               /* next day! */
               ray->utc -=23.0;
               ray->hour_mod = -(info->hour);
               CHILIncrDate(&(ray->day),&(ray->month),&(ray->year));
            }
            else {
               ray->utc +=1;
               ray->hour_mod++;
            }
         }
         ray->elevation=90.0;
         ray->azimuth=0.0;
      } 
      else {
         ray->utc=0.0;
     }
   
/* Now read the ray data itself */
      if (swap_bytes) {
         fread(tmp_buffer, sizeof(short), nwords,chilfile->file);
         swab((void *) &tmp_buffer,(void *)  &ray_buffer, nwords * 2);
      } 
      else {
         fread(ray_buffer, sizeof(short), nwords, chilfile->file);
      }
      (chilfile->bytes_read) += nwords * 2;
      chilfile->position = START_OF_RAY;
   }
   else {
      if (nwords == END_OF_RASTER || nwords == RASTER_ABORTED) {
         CHILNextRaster(chilfile,info);
      }
      if (nwords == END_OF_SCAN) {
         chilfile->position=START_OF_IR;
      }
      return -1;
   }
   ray->correct_range=0;
   ray->nez=0.0;
   return 0;
}

int CHILReadRay(ChilboltonFile * chilfile,RasterInfo * info,Ray * ray) {
int status,i,j;
unsigned short int field,gate,day,month,year,first_ray=1;
float (*func) (short),utc_mod=0;

if (ray->avegates == 1 && ray->averays == 1) {
/* NO AVERAGING CASE */
#ifdef DEBUG
fprintf(stderr,"Reading unaveraged ray\n");
#endif
   if ((status=CHILReadRawRay(chilfile,info,ray)) == 0 ) {
#ifdef DEBUG
fprintf(stderr,"...got raw data\n");
#endif
     for (field=0;field < ray->nfields;field++) {
         func=info->field_function[field];
         for (gate=0;gate < ray->ngates;gate++) {
//fprintf(stderr,"%hu %hd %hu; ",field,(int) ray_buffer[gate * (ray->nfields) + field],gate);
           *(ray->data[field]+gate) = func(ray_buffer[gate * (ray->nfields) + field]);
//fprintf(stderr,"; ");
         }
      }
   }
   else {
      return status;
   }
}
else {
/* AVERAGING CASE - first set everything to zero*/
#ifdef DEBUG
fprintf(stderr,"Reading averaged ray\n");
#endif

   ray->sum_elevation=0;
   ray->sum_azimuth=0;
   ray->sum_utc=0;

/* Remember that utc is the *mean* utc of an averaged ray, so when that ray
spans midnight the `mean' date of the ray can be `yesterday' and unless care 
is taken the next ray will think it's yesterday too... */
   ray->utc=ray->raw_utc;
   ray->day=ray->raw_day;
   ray->month=ray->raw_month;
   ray->year=ray->raw_year;

   for (field=0;field < ray->ntotal;field++) {
      for (gate=0;gate < ray->ngates;gate++) {
/*           *(ray->data[field]+gate) = 0;*/
           *(ray->sum[field]+gate) = 0;
      }
   }
   for (i=0;i<ray->averays;i++) {
      if ((status=CHILReadRawRay(chilfile,info,ray)) != 0 ) {
         return status;
      }
      if (first_ray) {
         day=ray->day;
         month=ray->month;
         year=ray->year;
         first_ray=0;
      }
      else if (ray->last_utc>ray->utc) {
         utc_mod = 24.0;
      }
      ray->sum_elevation += ray->elevation;
      ray->sum_azimuth += ray->azimuth;
      ray->sum_utc += (ray->utc + utc_mod);
      for (field=0;field < ray->nfields;field++) {
         func=info->field_function[field];

         /* if Z or LDR, convert from log to linear before averging */
         if (ray->field_type[field]==ZH || ray->field_type[field]==LDR) {
            for (gate=0;gate < ray->ngates;gate++) {
               for (j=gate*ray->avegates;j<(gate+1)*ray->avegates;j++) {
                 *(ray->sum[field]+gate) += pow(10,(double) 0.1 * func(ray_buffer[j * (ray->nfields) + field]));
               }
            }
         }

         /* otherwise convert in native form */
         else {
            for (gate=0;gate < ray->ngates;gate++) {
               for (j=gate*ray->avegates;j<(gate+1)*ray->avegates;j++) {
                 *(ray->sum[field]+gate) += (double) func(ray_buffer[j * (ray->nfields) + field]);
               }
            }
         }
      }
      /* next we do the derived fields (if there are any) */
      for (field=ray->nfields;field<ray->ntotal;field++) {
         /* standard deviation of velocity is the only one we know what to do with... */
         if (ray->field_type[field] == V_STD) {
            double tmp_double;
            func=info->field_function[ray->field_info[field]];
            for (gate=0;gate < ray->ngates;gate++) {
               for (j=gate*ray->avegates;j<(gate+1)*ray->avegates;j++) {
                 tmp_double = (double) func(ray_buffer[j * (ray->nfields) + ray->field_info[field]]);
                 *(ray->sum[field]+gate) += (tmp_double * tmp_double);
               }
            }
         }
      }
   }

   ray->raw_utc=ray->utc;
   ray->raw_day=ray->day;
   ray->raw_month=ray->month;
   ray->raw_year=ray->year;

   /* Work out averaged values from summation */
   ray->elevation=ray->sum_elevation / ray->averays;
   ray->azimuth=ray->sum_azimuth / ray->averays;
   ray->utc=ray->sum_utc / ray->averays;
   if (ray->utc > 24) {
      CHILIncrDate(&day,&month,&year);
      ray->utc -= 24;
   }

   ray->day=day;
   ray->month=month;
   ray->year=year;

   for (field=0;field < ray->ntotal;field++) {
      if (ray->field_type[field]==ZH || ray->field_type[field]==LDR) {
         for (gate=0;gate < ray->ngates;gate++) {
           *(ray->data[field]+gate) = 10.0 * log10(*(ray->sum[field]+gate) / 
                  (ray->avegates * ray->averays));
         }
      }
      else if (ray->field_type[field]==V_STD) {
         /* This is a `derived' field */
#ifdef DEBUG
fprintf(stderr,"Finalising sigma-v\n");
#endif
         for (gate=0;gate < ray->ngates;gate++) {
            *(ray->data[field]+gate) = (float) sqrt((*(ray->sum[field]+gate) / (ray->avegates * ray->averays) )
            - (double) (*(ray->data[ray->field_info[field]]+gate) * *(ray->data[ray->field_info[field]]+gate)));
         }
#ifdef DEBUG
fprintf(stderr,"(%f,%lf,%lf,%f,%f)",*(ray->range+100),*(ray->sum[1]+100),*(ray->sum[2]+100),*(ray->data[1]+100),*(ray->data[2]+100));
#endif 
      }
      else {
         for (gate=0;gate < ray->ngates;gate++) {
           *(ray->data[field]+gate) = *(ray->sum[field]+gate) / 
                  (ray->avegates * ray->averays);
         }
      }
   }
}
//fprintf(stderr,"[%hu %hu] ",ray->day,info->day);
return 0;
}
int CHILNextRaster(ChilboltonFile * chilfile,RasterInfo * info) {

int pad_size;

//fprintf(stderr,"Jumping to next raster...\n");
pad_size= BLKSIZE - chilfile->bytes_read % BLKSIZE;
if (pad_size > 0) {
   fread(pad_block, sizeof(char), pad_size, chilfile->file);
}
chilfile->bytes_read = 0;
(chilfile->iraster)++;
chilfile->iscan = 0;
return (chilfile->position=START_OF_IR);
}

/*
 * removes pixels where velocity std is less than MIN_V_STD (ground clutter)
 * and velocity is near zero.
 */
int CHILSuppressClutter(Ray *ray,float nosigval) {
word i,j,iSV=0,iV=0;
/* find which parameter is std of v, if available */
while (ray->field_type[iSV] != V_STD && iSV<ray->ntotal-1)
   iSV++;
/* find which parameter is v, if available */
while (ray->field_type[iV] != V1 && ray->field_type[iV] != V2 && iV<ray->ntotal-1)
   iV++;

/* Then remove (most) ground clutter */
if (ray->field_type[iSV] == V_STD) {
   for (i=0;i<ray->ngates;i++) {
      if (*(ray->data[iSV]+i) < MIN_V_STD && *(ray->data[iSV]+i) != nosigval && 
               *(ray->data[iV]+i) > - MIN_V_STD && *(ray->data[iV]+i) < MIN_V_STD) {
         for (j=0;j<ray->ntotal;j++) {
            *(ray->data[j]+i)=nosigval;
         }
      }
   }
}


}

/* 
 * Simple cleanup, to be applied AFTER CHILSubtractNoise
 * removes `isolated' pixels in ray which gets rid of some of those annoying lines
 */
int CHILCleanRay(Ray *ray,float nosigval) {
word i,j;

/* Remove isolated pixels */
for (i=1;i<ray->ngates-1;i++) {
   if (*(ray->data[0]+i-1) == nosigval && *(ray->data[0]+i+1) == nosigval) {
      for (j=0;j<ray->ntotal;j++) {
         *(ray->data[j]+i)=nosigval;
      }
   }
}
return 0;
}

/* Simple calibration of the Zh field (and nez value) */
int CHILCalibrateZh(Ray *ray,float Zcal,float nosigval) {
word iZ=0,i=0,j;

/* find which parameter is Z */
while (ray->field_type[iZ] != ZH && iZ<ray->ntotal)
   iZ++;
if (ray->field_type[iZ] != ZH) {
#ifdef DEBUG
   fprintf(stderr,"CHILCalibrateZh: attempt to calibrate Zh when Zh not a parameter\n");
#endif
   return 2;
}
for (i=0;i<ray->ngates;i++) {
   if (*(ray->data[iZ]+i) != nosigval) 
      (*(ray->data[iZ]+i)) += Zcal;
}
ray->nez += Zcal;
return 0;
}
/*
 * Simple noise subtraction, using the far gates to get the noise level.
 * Noise is subtracted from Z only; the remaining fields are masked only.
 * If using CHILCleanRay too, call this function FIRST
 */
int CHILSubtractNoise(Ray *ray,int noisegates,float factor,float nosigval) {
word iZ=0,i=0,j;
double sum_noiseP=0.0,sum_noiseP2=0.0,noiseP=0.0,P,std_noiseP;
float sfloor[MAX_FIELDS];

if (noisegates <1 || noisegates > ray->ngates) {
   fprintf(stderr,"CHILSubtractNoise: %d is an inappropriate number of noise gates\n",noisegates);
   return 1;
}

/* find which parameter is Z */
while (ray->field_type[iZ] != ZH && iZ<ray->ntotal)
   iZ++;
if (ray->field_type[iZ] != ZH) {
   fprintf(stderr,"CHILSubtractNoise: attempt to subtract noise when Zh not a parameter\n");
   return 2;
}

/* Set the rejection thresholds for each parameter relative to that for Z */
/* in multiplicative units */
for (j=0;j<ray->ntotal;j++) {
   if (ray->field_type[j] == V1 || ray->field_type[j] == V2) {
      sfloor[j]=pow(10,0.1*ray->v_floor);
   }
   else if (ray->field_type[j] == V_STD) {
      sfloor[j]=pow(10,0.1*ray->sigma_v_floor);
   }
   else
      sfloor[j]=1.0;
}


/* estimate the noise level from the top gates */
if (ray->range_normalised) {
   for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP = (pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      sum_noiseP += noiseP;
      sum_noiseP2 += (noiseP * noiseP);
   }
}
else {
   for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP = (pow(10,0.1* (*(ray->data[iZ]+i))));
      sum_noiseP += noiseP;
      sum_noiseP2 += (noiseP * noiseP);
   }
}

ray->nez=10.0*log10(noiseP = sum_noiseP / noisegates);
std_noiseP = sqrt(sum_noiseP2/noisegates - noiseP * noiseP);

/* Subtract noise from Zh and mask other parameters*/
if (ray->range_normalised) {
   for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      for (j=0;j<ray->ntotal;j++) {
         if (P < sfloor[j]*noiseP + factor*std_noiseP) {
            (*(ray->data[j]+i))=nosigval;
         }
      }
      if ((*(ray->data[iZ]+i)) != nosigval) {
         (*(ray->data[iZ]+i))=(float) (10.0*log10((P-noiseP)*ray->range[i]*ray->range[i]));
      }
   }
}
else {
   for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))));
      for (j=0;j<ray->ntotal;j++) {
         if (P < sfloor[j]*noiseP + factor*std_noiseP) {
            (*(ray->data[j]+i))=nosigval;
         }
      }
      if (*(ray->data[iZ]+i) != nosigval) {
         (*(ray->data[iZ]+i))=(float) (10.0*log10(P-noiseP));
      }
   }
}
return 0;
}

/*
 * More advanced noise subtraction. 
 * Assumes there is interference from some kind of oscillator and uses noisegates
 * gates from the end of the ray to characterise it.
 * Noise is subtracted from Z only; the remaining fields are masked only.
 * If using CHILCleanRay too, call this function FIRST
 */
int CHILSubtractOscillatorNoise(Ray *ray,int noisegates,float factor,float nosigval) {
word iZ=0,i=0,j;
double sum_diff_noiseP2=0.0,noiseP[1023],P,std_noiseP,sum_noiseP=0;
float sfloor[MAX_FIELDS];

if (noisegates <1 || noisegates > ray->ngates) {
   fprintf(stderr,"CHILSubtractNoise: %d is an inappropriate number of noise gates\n",noisegates);
   return 1;
}

/* find which parameter is Z */
while (ray->field_type[iZ] != ZH && iZ<ray->ntotal)
   iZ++;
if (ray->field_type[iZ] != ZH) {
   fprintf(stderr,"CHILSubtractNoise: attempt to subtract noise when Zh not a parameter\n");
   return 2;
}

/* Set the rejection thresholds for each parameter relative to that for Z */
/* in multiplicative units */
for (j=0;j<ray->ntotal;j++) {
   if (ray->field_type[j] == V1 || ray->field_type[j] == V2) {
      sfloor[j]=pow(10,0.1*ray->v_floor);
   }
   else if (ray->field_type[j] == V_STD) {
      sfloor[j]=pow(10,0.1*ray->sigma_v_floor);
   }
   else
      sfloor[j]=1.0;
}

/* get the noise level from the top gates */
if (ray->range_normalised) {
   for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP[i] = (pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      sum_noiseP += noiseP[i];
   }
}
else {
   for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP[i] = (pow(10,0.1* (*(ray->data[iZ]+i))));
      sum_noiseP += noiseP[i];
   }
}

ray->nez=10.*log10(sum_noiseP/noisegates);

/* estimate the standard deviation from the difference between adjacent gates */
for (i=ray->ngates-noisegates+1;i<ray->ngates;i++) {
   sum_diff_noiseP2 += (noiseP[i]-noiseP[i-1])*(noiseP[i]-noiseP[i-1]);
}
std_noiseP=sqrt(sum_diff_noiseP2/(noisegates-1));

/* replicate the noise pattern down through the other gates */
for (i=ray->ngates-noisegates-1;i>=0;i--) {
   noiseP[i]=noiseP[i+noisegates];
}

/* Subtract noise from Zh and mask other parameters*/
if (ray->range_normalised) {
   for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      for (j=0;j<ray->ntotal;j++) {
         if (P < sfloor[j]*noiseP[i] + factor*std_noiseP) {
            (*(ray->data[j]+i))=nosigval;
         }
      }
      if (*(ray->data[iZ]+i) != nosigval) {
         (*(ray->data[iZ]+i))=(float) (10.0*log10((P-noiseP[i])*ray->range[i]*ray->range[i]));
      }
   }
}
else {
   for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))));
      for (j=0;j<ray->ntotal;j++) {
         if (P < sfloor[j]*noiseP[i] + factor*std_noiseP) {
            (*(ray->data[j]+i))=nosigval;
         }
      }
      if (*(ray->data[iZ]+i) != nosigval) {
         (*(ray->data[iZ]+i))=(float) (10.0*log10(P-noiseP[i]));
//         fprintf(stderr,"(%e %e) ",(*(ray->data[iZ]+i)),(float) (10.0*log10(P-noiseP[i])));
      }
   }
}
return 0;
}

/* The following function performs range correction;
This involves correcting Z according to what's in the range_offset
member of RasterInfo */
int CHILRangeCorrection(RasterInfo *info,Ray *ray,float nosigval) {
word iZ=0,i;
float range_offset;

range_offset=0.001*ray->range_offset;

if (ray->correct_range) {
   fprintf(stderr,"Range normalisation already performed\n");
   return 2;
}

if (range_offset==0) {
   fprintf(stderr,"Attempt to range normalise when range offset is unknown\n");
   return 1;
}

/* find which parameter is Z */
while (ray->field_type[iZ] != ZH && iZ<ray->ntotal)
   iZ++;

if (ray->field_type[iZ] == ZH) {
   if (ray->range_normalised) {
      for (i=0;i<ray->ngates;i++) {
         if (ray->range[i]+range_offset > 0.0 && 
                                                      *(ray->data[iZ]+i) != nosigval) {
            *(ray->data[iZ]+i) +=
                  20.0*log10((ray->range[i]+range_offset)/(ray->range[i]));
         }
         else {
            *(ray->data[iZ]+i)=nosigval;
         }
      }
   }
   else {
      for (i=0;i<ray->ngates;i++) {
         if (ray->range[i]+range_offset > 0.0 && 
                                                         *(ray->data[iZ]+i) != nosigval) {
            *(ray->data[iZ]+i) +=
                  20.0*log10(ray->range[i]+range_offset);
         }
         else
            *(ray->data[iZ]+i)=nosigval;
      }
   }
}

ray->correct_range=1;
return 0;
}
int CHILPrintFileInformation(ChilboltonFile * chilfile) {
int i;

fprintf(stderr,"File %s:\n",chilfile->name);
fprintf(stderr,"   Position: ");
switch (chilfile->position) {
   case FILE_CLOSED:
      fprintf(stderr,"file closed\n");
      break;
   case START_OF_IR:
      fprintf(stderr,"start of information record\n");
      break;
   case START_OF_DATA:
      fprintf(stderr,"start of data\n");
      break;
   case START_OF_RAY:
      fprintf(stderr,"start of ray\n");
      break;
   case START_OF_FILE:
      fprintf(stderr,"start of file\n");
      break;
   default:
      fprintf(stderr,"unknown\n");
}
if (chilfile->is_raw) {
   fprintf(stderr,"   File is raw\n");
} 
else {
//   fprintf(stderr,"   File is PROCESSED (MSB)\n");
   if (chilfile->summary_exists) {
      fprintf(stderr,"   Number of rasters: %hu\n",chilfile->nrasters);
/*
      fprintf(stderr,"   Raster numbers:");
      for (i=0;i<chilfile->nrasters;i++) {
         fprintf(stderr," %hu",chilfile->raster[i]);
      };
*/
   }
   else
      fprintf(stderr,"   No summary file found\n");
   fprintf(stderr,"   Current raster index (0 based): %hu\n",chilfile->iraster);

}
fprintf(stderr,"   Current scan index (0 based):   %hu\n",chilfile->iscan);
};
int CHILPrintInformationRecord(RasterInfo * info) {
uword i;
char system[10];

fprintf(stderr,"Raster %hd:\n",info->raster);
CHILSystemName(info,system);
//fprintf(stderr,"   System: %s (ID %hu)\n",system_names[info->system_id],info->system_id);
fprintf(stderr,"   System: %s (ID %hu)\n",system,info->system_id);
fprintf(stderr,"   Frequency: %f GHz\n",info->frequency);
fprintf(stderr,"   Date: %hu/%hu/%hu\n",info->day,info->month,info->year);
fprintf(stderr,"   Start time: %02hu:%02hu:%02hu\n",info->hour,info->minute,info->second);
if (info->cloud_data)
   fprintf(stderr,"   Cloud data\n");
else
   fprintf(stderr,"   Not cloud data\n");

fprintf(stderr,"   Number of scans: %hu\n",info->nscans);
fprintf(stderr,"   Number of fields: %hu\n",info->nfields);
fprintf(stderr,"   Time stamping on? %hu\n",info->time_stamping);
fprintf(stderr,"   Minimum gate: %hu\n",info->min_gate);
fprintf(stderr,"   Maximum gate: %hu\n",info->max_gate);
fprintf(stderr,"   Gate spacing: %hd m\n",info->gate_spacing);
if (info->range_offset != 0)
   fprintf(stderr,"   Range offset: %hd m\n",info->range_offset);
}

int CHILPrintFieldInformation(Ray * ray) {
   char fielddesc[64];
   int desclen,i;

   fprintf(stderr,"Data acquisition setup:\n");
   fprintf(stderr,"   Averaging %hu gates\n",ray->avegates);
   fprintf(stderr,"   Averaging %hu rays\n",ray->averays);
   fprintf(stderr,"   Fields:\n");
   for (i=0;i< ray->ntotal;i++) {
      desclen=strlen(CHILFieldName(ray,i,fielddesc));
      strcpy(fielddesc+desclen,", ");
      desclen=desclen+2+strlen(CHILFieldDescription(ray,i,fielddesc+desclen+2));
      strcpy(fielddesc+desclen," (");
      desclen=desclen+2+strlen(CHILFieldUnits(ray,i,fielddesc+desclen+2));
      strcpy(fielddesc+desclen,")");
      fprintf(stderr,"      %s\n",fielddesc);
   }
   return 0;
}

int CHILPrintRayInformation(Ray * ray) {
fprintf(stderr,"Ray %hu:\n",ray->index);
fprintf(stderr,"   Number of gates: %hu\n",ray->ngates);
fprintf(stderr,"   Number of fields: %hu\n",ray->nfields);
fprintf(stderr,"   Decimal hours: %f UTC\n",ray->utc);
fprintf(stderr,"   Elevation: %f\n",ray->elevation);
fprintf(stderr,"   Azimuth: %f\n",ray->azimuth);

};
#define RAL_BAD -999.0
float CHILzh(short x) {
    return (((float) x) * ZH_SCALING_93);
};
float CHILraw(short x) {
    return ((float) x);
};
float CHILzh_highprecision(short x) {
    return (((float) x) * ZH_SCALING_HP);
};
float CHILzdr(short x) {
    return (((float) x) * ZDR_SCALING_93);
};
float CHILzdr_highprecision(short x) {
    return (((float) x) * ZDR_SCALING_HP);
};
float CHILldr(short x) {
    return (((float) x) * LDR_SCALING_93);
};
float CHILphidp_old(short x) {
   return (0.5 * atan(((float) x) * PHASE_SCALING_93));
};
float CHILphidp(short x) {
    return (((float) x) * PHASE_SCALING_93);
};
float CHILvel_old(short x) {
   return (-2.3679155 * atan(((float) x) * PHASE_SCALING_93));
};
float CHILvel(short x) {
   return (((float) x) * PHASE_SCALING_93);
};


