%
% process3ghz.m
%
% Performs processing to a 3 GHz reflectivity scan, including:
%   removing noise and noise effects from Zh, Zdr and Ldr
%   averages rays/gates if required
%   screens out data for which Ldr is above a certain threshold (ground clutter)
%   `despeckles' the data if required
%  
% WARNING: some default noise levels for low elevations (when they cannot
% obtained from the last gates in the ray) are hard-wired - may need to be 
% changed if ever the hardware configuration changes.


[ng nr] = size(Zh);

% Existence of Zdr and Ldr?
if ~exist('iszdr'), iszdr=(exist('Zdr')>0);end
if ~exist('isldr'), isldr=(exist('Ldr')>0);end

% Despeckling? Default YES
if ~exist('despeckling'), despeckling=1;despeckle_mode='weak';end


% Threshold for LDR screening of ground clutter
% use number greater than around 10 for no screening
if ~exist('LDRthreshold'),LDRthreshold=-10;end

% Calibration figures for the 3 GHz
if ~exist('Zcal'),Zcal = 4;end
disp(['Calibration figure for Z: ' num2str(Zcal) ' dB'])
if iszdr
   if ~exist('ZDRcal'),ZDRcal = 1.5;end
   disp(['Calibration figure for ZDR: ' num2str(ZDRcal) ' dB'])
end

% Extent of noise rejection
if ~exist('tfactor'),tfactor = 3.0;end

% Default noise level (for low elevations)
if ~exist('ZHnoise_normal'),ZHnoise_normal=-41.7+Zcal;end
if ~exist('Zstdn_normal'),Zstdn_normal=0.258;end
if iszdr,if ~exist('ZVnoise_normal'),ZVnoise_normal=-43.1+Zcal-ZDRcal;end;end
if isldr
   if ~exist('ZHVnoise_normal'),ZHVnoise_normal=-43.1+Zcal;end;
   if ~exist('ZHVstdn_normal'),ZHVstdn_normal=0.28;end;
end

% Number of rays and gates to average (default 1)
if ~exist('sr'),sr=1;end
if ~exist('sg'),sg=1;end

% First gate at which noise is estimated
noisegate = (ng-20)./sg;

% Perform averaging if required and apply Z and Zdr calibration figure
if sr > 1 | sg > 1
   [elevs,ranges,ZH] = CRFaverageZ(elev,range,Zh+Zcal,sr,sg);
   if iszdr,[elevs,ranges,ZDR] = CRFaverageZ(elev,range,Zdr+ZDRcal,sr,sg);end
   if isldr,[elevs,ranges,LDR] = CRFaverageZ(elev,range,Ldr,sr,sg);end
   nr = fix(nr/sr); 
   ng = fix(ng/sg);
   azims=mean(reshape(azim(1:nr*sr),sr,nr));
else
   ZH=Zh+Zcal;
   if iszdr,ZDR=Zdr+ZDRcal;end
   if isldr,LDR=Ldr;end
   elevs=elev;
   ranges=range;
   azims=azim;
end

if iszdr,ZV=ZH-ZDR;end
if isldr,ZHV=ZH+LDR;end

% calculate x, y, z and r for plotting (all in km):
% for RHIs, use pcolor(r,z,ZH);
% for PPIs, use pcolor(x,y,ZH);
r = cos(elevs'*pi/180)*ranges;
x = -(cos(azims'*pi/180).*cos(elevs'*pi/180))*ranges;
y = (sin(azims'*pi/180).*cos(elevs'*pi/180))*ranges;
z = sin(elevs'*pi/180)*ranges;
switch sweepm
   case 0,
      disp('Scan type: calibration')
   case 1,
      disp('Scan type: PPI')
   case 2,
      disp('Scan type: coplane')
   case 3,
      disp('Scan type: RHI')
   case 4,
      disp('Scan type: vertical')
   case 5,
      disp('Scan type: dwell')
   case 6,
      disp('Scan type: manual')
   case 7,
      disp('Scan type: idle')
end

Zbias = zeros(1,ng);
R = 20.*log10(ranges')*ones(1,nr);
RN = 20.*log10(ranges([noisegate:ng])')*ones(1,nr);
RNB = 20.*log10(ranges([noisegate:ng])' - Zbias([noisegate:ng])')*ones(1,nr);

% NOISE SUBTRACTION AND THRESHOLDING FOR HORIZONTAL POLARISATION
disp('Processing horizontal polarisation')
noise = 10.*log10(mean(10.^(0.1*(ZH([noisegate:ng],:) - RNB))));
stdn = std(ZH([noisegate:ng],:) - RNB);
index=find(elevs<20);
noise(index)=ZHnoise_normal;
stdn(index)=Zstdn_normal./sqrt(sr.*sg);
noiseZH=noise;
stdnZH=stdn;

ZHn = R + ones(ng,1)*noise + Zbias'*ones(1,nr);
index = find((ZH-ZHn-tfactor*ones(ng,1)*stdn) < 0.0);
ZH(index) = NaN*ones(size(index));
ZH = 10.*log10(10.^(0.1*ZH)-10.^(0.1*ZHn));

if iszdr
% NOISE SUBTRACTION AND THRESHOLDING FOR VERTICAL POLARISATION
disp('Processing vertical polarisation')
noise = 10.*log10(mean(10.^(0.1*(ZV([noisegate:ng],:) - RNB))));
index=find(elevs<20);
noise(index)=ZVnoise_normal;

ZVn = R + ones(ng,1)*noise + Zbias'*ones(1,nr);
index = find((ZV-ZVn-tfactor*ones(ng,1)*stdn) < 0.0);
ZV(index) = NaN*ones(size(index));
ZV = 10.*log10(10.^(0.1*ZV)-10.^(0.1*ZVn));
ZDR = ZH - ZV;
end

if isldr
% NOISE SUBTRACTION AND THRESHOLDING FOR CROSS POLAR CHANNEL
disp('Processing cross-polar channel')
noise = 10.*log10(mean(10.^(0.1*(ZHV([noisegate:ng],:) - RN))));
stdn = std(ZHV([noisegate:ng],:) - RN);
index=find(elevs<20);
noise(index)=ZHVnoise_normal;
stdn(index)=ZHVstdn_normal./sqrt(sr.*sg);
ZHVn =  R + ones(ng,1)*noise;
index = find((ZHV - ZHVn - tfactor*ones(ng,1)*stdn) < 0.0);
ZHV(index) = NaN*ones(size(index));
ZHV = 10.*log10(10.^(0.1*ZHV)-10.^(0.1*ZHVn));
LDR = ZHV - ZH;
end

% Screen out ground clutter using LDR
if isldr
   disp(['Screening out data for which LDR > ' num2str(LDRthreshold) ' dB'])
   index=find(LDR>LDRthreshold);
   ZH(index)=NaN;
   if iszdr
      ZV(index)=NaN;
      ZDR(index)=NaN;
   end
   LDR(index)=NaN;
end

if despeckling
   ZH=CRFdespeckle(ZH,despeckle_mode);
   if iszdr
      ZV=CRFdespeckle(ZV,despeckle_mode);
      ZDR=CRFdespeckle(ZDR,despeckle_mode);
   end
   if isldr,LDR=CRFdespeckle(LDR,despeckle_mode);end
end


clear ZHVn ZHn ZVn index 
%noise stdn ZHV 
clear R RN RNB Zbias

% Do time-stamping 
if exist('centisec')
   Rutc=minutes./60.0+centisec./360000.0;
   disp('Time stamping is ON')
end





